
import abc
import asyncio
import io
import logging
import pathlib
import tarfile
from collections import defaultdict
from concurrent import futures
from functools import cached_property
from typing import AsyncIterator, List, Optional, Tuple, Type

import aiohttp

import abstracts

from aio.api import nist
from aio.core import event
from aio.core.functional import (
    async_property,
    AwaitableGenerator,
    qdict)

from envoy.base import utils
from envoy.dependency.check import abstract, typing


logger = logging.getLogger(__name__)


@abstracts.implementer(event.IReactive)
class ADependencyCVEs(event.AReactive, metaclass=abstracts.Abstraction):

    def __init__(
            self,
            dependencies: Tuple["abstract.ADependency", ...],
            config_path: Optional[str] = None,
            preloaded_cve_data: Optional[str] = None,
            session: Optional[aiohttp.ClientSession] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        self.dependencies = dependencies
        self._config_path = config_path
        self._preloaded_cve_data = preloaded_cve_data
        self._session = session
        self._loop = loop
        self._pool = pool

    @cached_property
    def config(self) -> "typing.CVEConfigDict":
        """CVE scan config - a combination of defaults and config defined
        in provided config file
        """
        return (
            utils.typed(dict, utils.from_yaml(self.config_path))
            if self.config_path
            else {})

    @property
    def config_path(self) -> Optional[pathlib.Path]:
        """Path to CVE scan config file."""
        return (
            pathlib.Path(self._config_path)
            if self._config_path
            else None)

    @property  # type:ignore
    @abstracts.interfacemethod
    def cpe_class(self) -> Type["nist.ACPE"]:
        """CPE class."""
        raise NotImplementedError

    @cached_property
    def cpe_revmap(self) -> "nist.typing.CPERevmapDict":
        return defaultdict(set)

    @property  # type:ignore
    @abstracts.interfacemethod
    def cve_class(self) -> Type["abstract.ADependencyCVE"]:
        """CVE class."""
        raise NotImplementedError

    @property
    def cve_fields(self):
        return qdict(
            score="impact/baseMetricV3/cvssV3/baseScore",
            severity="impact/baseMetricV3/cvssV3/baseSeverity",
            description="cve/description/description_data/0/value",
            last_modified_date="lastModifiedDate")

    @cached_property
    def cves(self) -> "nist.typing.CVEDict":
        return {}

    @async_property(cache=True)
    async def data(self) -> "nist.typing.CVEDataTuple":
        if not await self.loader:
            await AwaitableGenerator(self.downloads)
        return self.cves, self.cpe_revmap

    @async_property
    async def downloads(self) -> AsyncIterator[str]:
        """Yield and optionally download from NIST CVE urls.

        If data is already loaded will just yield the urls, otherwise
        downloads and captures the data.
        """
        if await self.loader:
            for url in self.nist_downloader.urls:
                yield url
            return
        with self.loader:
            async for url in self.nist_downloader:
                yield url
            for id, cve in self.nist_downloader.cves.items():
                self.cves[id] = self.cve_class(cve, self.cpe_class)
            self.cpe_revmap.update(self.nist_downloader.cpe_revmap)

    @property
    @abc.abstractmethod
    def ignored_cves(self) -> List[str]:
        """List of CVEs to ignore, taken from config file."""
        return self.config.get("ignored_cves", [])

    @cached_property
    def loader(self) -> event.ILoader:
        return event.Loader()

    @cached_property
    def nist_downloader(self):
        return self.nist_downloader_class(
            self.tracked_cpes,
            cve_fields=self.cve_fields,
            ignored_cves=self.ignored_cves,
            cve_data=self.preloaded_cve_data,
            since=self.scan_year_start,
            pool=self.pool,
            session=self.session)

    @property  # type:ignore
    @abstracts.interfacemethod
    def nist_downloader_class(self) -> Type["nist.abstract.ANISTDownloader"]:
        """NIST downloader class."""
        raise NotImplementedError

    @property
    def preloaded_cve_data(self) -> Optional[str]:
        return self._preloaded_cve_data

    @property
    def scan_year_start(self) -> Optional[int]:
        """Inclusive start year to scan from."""
        return self.config.get("start_year")

    @cached_property
    def session(self) -> aiohttp.ClientSession:
        """HTTP client session."""
        return self._session or aiohttp.ClientSession()

    @cached_property
    def tracked_cpes(self) -> "nist.typing.TrackedCPEDict":
        """Dict of tracked CPE <> dependency."""
        return {
            v.cpe: self.cpe_filter_dict(v)
            for v
            in self.dependencies
            if v.cpe}

    def cpe_filter_dict(
            self,
            dependency: "abstract.ADependency") -> (
                "nist.typing.TrackedCPEFilterDict"):
        """Provide filter args to filter out non-matching CPEs."""
        return dict(
            version=dependency.release_version,
            date=dependency.release_date)

    async def dependency_check(
            self,
            dep: "abstract.ADependency") -> AsyncIterator[
                "abstract.ADependencyCVE"]:
        """Check for relevant CVEs for a given dep."""
        if not dep.cpe:
            return
        cves, cpe_revmap = await self.data
        cpe = self.cpe_class.from_string(dep.cpe).vendor_normalized
        for cpe_cve in sorted(cpe_revmap.get(cpe, [])):
            yield cves[cpe_cve]

    async def download_cves(self, path: str) -> None:
        with tarfile.open(path, "w") as tar:
            for url in self.nist_downloader.urls:
                response = await self.nist_downloader.download(url)
                data = io.BytesIO(await response.read())
                info = tarfile.TarInfo(name=url.split("/")[-1])
                info.size = len(data.getvalue())
                tar.addfile(info, fileobj=data)




import asyncio
import hashlib
import logging
import time
from concurrent import futures
from datetime import datetime
from functools import cached_property
from typing import Optional

from packaging import version

import aiohttp

import gidgethub

import abstracts

from aio.api import github as _github
from aio.core import event
from aio.core.functional import async_property

from envoy.base import utils
from envoy.dependency.check import exceptions


logger = logging.getLogger(__name__)

MIN_DATA_SIZE_TO_HASH_IN_PROC = 1000000


@abstracts.implementer(event.IExecutive)
class ADependencyGithubRelease(
        event.AExecutive,
        metaclass=abstracts.Abstraction):
    """Github release associated with a dependency."""

    @classmethod
    def hash_file_data(cls, data: bytes) -> str:
        file_hash = hashlib.sha256()
        file_hash.update(data)
        return str(file_hash.hexdigest())

    def __init__(
            self,
            repo: _github.IGithubRepo,
            version: str,
            asset_url: Optional[str] = None,
            release: Optional[_github.IGithubRelease] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        self.repo = repo
        self.asset_url = asset_url
        self._version = version
        self._release = release
        self._pool = pool
        self._loop = loop

    @async_property(cache=True)
    async def commit(self) -> Optional[_github.IGithubCommit]:
        """Github commit for this release."""
        try:
            return await self.repo.commit(self.tag_name)  # type:ignore
        except gidgethub.BadRequest as e:
            if e.args[0] == "Not Found":
                return None
            raise e

    @async_property(cache=True)
    async def date(self) -> Optional[str]:
        """UTC date of this release."""
        # TODO(phlax): add tests for this and related fun
        return (
            utils.dt_to_utc_isoformat(timestamp)
            if (timestamp := await self.timestamp)
            else None)

    @property
    def github(self) -> _github.IGithubAPI:
        return self.repo.github

    @property
    def min_data_size_to_hash_in_proc(self) -> int:
        return MIN_DATA_SIZE_TO_HASH_IN_PROC

    @async_property(cache=True)
    async def release(self) -> Optional[_github.IGithubRelease]:
        """Github release."""
        if self._release:
            return self._release
        try:
            return await self.repo.release(self.tag_name)  # type:ignore
        except gidgethub.BadRequest as e:
            if e.args[0] == "Not Found":
                return None
            raise e

    @property
    def session(self) -> aiohttp.ClientSession:
        return self.github.session

    @async_property(cache=True)
    async def sha(self) -> str:
        """Release SHA."""
        if not self.asset_url:
            raise exceptions.NoReleaseAssetError(
                f"Cannot check sha for {self.__class__.__name__} "
                "with no `asset_url`")
        response = await self.session.get(self.asset_url)
        logger.debug(f"SHA download: {self.asset_url}")
        return await self._hash_file_data(await response.read())

    @async_property(cache=True)
    async def tag(self) -> Optional[_github.IGithubTag]:
        """Github tag."""
        try:
            return await self.repo.tag(self.tag_name)  # type:ignore
        except (gidgethub.BadRequest, _github.exceptions.TagNotFound) as e:
            do_raise = (
                isinstance(e, gidgethub.BadRequest)
                and e.args[0] != "Not Found")
            if do_raise:
                raise e
            return None

    @property
    def tag_name(self) -> str:
        """Github tag name."""
        return self._version

    @cached_property
    def tagged(self) -> bool:
        """Flag to indicate whether this release has a name."""
        return not utils.is_sha(self.tag_name)

    @async_property(cache=True)
    async def timestamp(self) -> Optional[datetime]:
        """Timestamp of this release."""
        return (
            await self.timestamp_tag
            if self.tagged
            else await self.timestamp_commit)

    @async_property(cache=True)
    async def timestamp_commit(self) -> Optional[datetime]:
        """Timestamp of the commit of this release."""
        return (
            commit.timestamp
            if (commit := await self.commit)
            else None)

    @async_property(cache=True)
    async def timestamp_tag(self) -> Optional[datetime]:
        """Timestamp of this release, resolved from the release, tag, or
        commit, in that order."""
        if release := await self.release:
            return release.published_at
        if await self.tag:
            return (await (await self.tag).commit).timestamp
        # Its not clear why this is required - it seems to be legacy tags
        # in this case fetching the `tag_name` as a commit seems to work.
        return await self.timestamp_commit

    @cached_property
    def version(self) -> Optional[version.Version]:
        """Semantic version of this release."""
        try:
            return version.parse(self.tag_name)
        except version.InvalidVersion:
            return None

    def should_hash_in_proc(self, data: bytes) -> bool:
        """Conditionally generate SHA hashes based on incoming data."""
        return len(data) > self.min_data_size_to_hash_in_proc

    async def _hash_file_data(self, data: bytes) -> str:
        hash_in_proc = self.should_hash_in_proc(data)
        start = time.perf_counter()
        sha = (
            await self.execute(self.hash_file_data, data)
            if hash_in_proc
            else self.hash_file_data(data))
        logger.debug(
            "SHA parsed in {:.3f}s{}: {}" .format(
                time.perf_counter() - start,
                (" (fork: True)"
                 if hash_in_proc
                 else ""),
                self.asset_url))
        return sha




from functools import cached_property
from typing import Dict, Optional, Tuple

from packaging import version

import abstracts

from aio.api import github as _github


GITHUB_REPO_LOCATION = "envoyproxy/envoy"
LABELS = ("dependencies", "area/build", "no stalebot")
BODY_TPL = """
Package Name: {dep}
Current Version: {dep.github_version_name}@{release_date}
Available Version: {newer_release.tag_name}@{newer_release_date}
Upstream releases: https://github.com/{dep.release.repo.name}/releases
"""
CLOSING_TPL = """
New version is available for this package
New Version: {newer_release.tag_name}@{newer_release_date}
Upstream releases: https://github.com/{full_name}/releases
New Issue Link: https://github.com/{repo_location}/issues/{number}
"""
TITLE_PREFIX = "Newer release available"
TITLE_RE_TPL = r"{title_prefix} [`]?([\w\-\.]+)[`]?: ([\w\-\.]+)"
TITLE_TPL = (
    "{title_prefix} `{dep.id}`: {newer_release.tag_name} "
    "(current: {dep.github_version_name})")


@abstracts.implementer(_github.IGithubTrackedIssue)
class AGithubDependencyReleaseIssue(
        _github.AGithubTrackedIssue,
        metaclass=abstracts.Abstraction):
    """Github issue associated with a dependency."""

    @property
    def parse_vars(self) -> Tuple[str, ...]:
        return ("key", "version")

    @cached_property
    def version(self) -> Optional[version.Version]:
        """Parsed dependency version of an issue."""
        try:
            return (
                version.parse(self.parsed["version"])
                if "version" in self.parsed
                else None)
        except version.InvalidVersion:
            return None

    async def close_old(
            self,
            old_issue: "AGithubDependencyReleaseIssue",
            **kwargs) -> None:
        # TODO: reassign any users old -> new issue
        newer_release = await kwargs["dep"].newer_release
        await old_issue.comment(
            self.closing_tpl.format(
                newer_release=newer_release,
                newer_release_date=await newer_release.date,
                full_name=kwargs["dep"].repo.name,
                repo_location=self.repo_name,
                number=self.number))
        await old_issue.close()


@abstracts.implementer(_github.IGithubTrackedIssues)
class AGithubDependencyReleaseIssues(
        _github.AGithubTrackedIssues,
        metaclass=abstracts.Abstraction):
    """Github issues associated with released dependencies."""

    @property
    def body_tpl(self) -> str:
        return BODY_TPL

    @property
    def closing_tpl(self) -> str:
        return CLOSING_TPL

    @property
    def issue_author(self) -> str:
        return super().issue_author

    @property
    def issues_search_tpl(self) -> str:
        return super().issues_search_tpl

    @property
    def labels(self) -> Tuple[str, ...]:
        return LABELS

    @property
    def repo_name(self) -> str:
        return GITHUB_REPO_LOCATION

    @property
    def title_prefix(self) -> str:
        return TITLE_PREFIX

    @property
    def title_re_tpl(self) -> str:
        return TITLE_RE_TPL

    @property
    def title_tpl(self) -> str:
        return TITLE_TPL

    async def issue_body(self, **kwargs) -> str:
        newer_release = await kwargs["dep"].newer_release
        return self.body_tpl.format(
            dep=kwargs["dep"],
            newer_release=newer_release,
            newer_release_date=await newer_release.date,
            release_date=await kwargs["dep"].release.date)

    async def issue_title(self, **kwargs) -> str:
        return self.title_tpl.format(
            dep=kwargs["dep"],
            title_prefix=self.title_prefix,
            newer_release=await kwargs["dep"].newer_release)

    def track_issue(  # type:ignore[override]
            self,
            issues: Dict[str, AGithubDependencyReleaseIssue],
            issue: AGithubDependencyReleaseIssue) -> bool:
        if issue.key not in issues:
            return True
        if not (existing_issue_version := issues[issue.key].version):
            return bool(issue.version)
        if not issue.version:
            return bool(existing_issue_version)
        return issue.version > existing_issue_version




from .checker import ADependencyChecker
from .cves import (
    ADependencyCVE,
    ADependencyCVEs)
from .dependency import ADependency
from .issues import (
    AGithubDependencyReleaseIssue,
    AGithubDependencyReleaseIssues)
from .release import ADependencyGithubRelease


__all__ = (
    "ADependency",
    "ADependencyChecker",
    "ADependencyCVE",
    "ADependencyCVEs",
    "ADependencyGithubRelease",
    "AGithubDependencyReleaseIssue",
    "AGithubDependencyReleaseIssues")



"""Abstract CVE."""

import textwrap
from datetime import date
from functools import cached_property
from typing import Set, Type

import jinja2

import abstracts

from envoy.dependency.check import abstract, exceptions, typing

from aio.api import nist


CVE_FAIL_TPL = """
  {{cve.id}} ({{dependency.id}}@{{dependency.display_version}})
  CVSS v3 score: {{cve.score}}
  Severity: {{cve.severity}}
  Published date: {{cve.published_date}}
  Last modified date: {{cve.last_modified_date}}
  Description: {{cve.formatted_description}}
  Affected CPEs:
{%- for cpe in cve.cpes %}
  - {{cpe}}
{% endfor %}
"""


class ADependencyCVE(metaclass=abstracts.Abstraction):

    def __init__(
            self,
            cve_data: "typing.DependencyCVEItemDict",
            cpe_class: Type["nist.ACPE"]) -> None:
        self.cve_data = cve_data
        self.cpe_class = cpe_class

    def __gt__(self, other) -> bool:
        return self.id > other.id

    @cached_property
    def cpes(self) -> Set["nist.ACPE"]:
        """Associated CPEs."""
        return set(self.cpe_class(**cpe) for cpe in self.cve_data["cpes"])

    @property
    def description(self) -> str:
        """CVE description."""
        return self.cve_data["description"]

    @cached_property
    def fail_template(self) -> jinja2.Template:
        """Jinja2 template for rendering a failing CVE match."""
        return jinja2.Template(self.fail_tpl)

    @property
    def fail_tpl(self) -> str:
        """Template string, used for rendering with Jinja2."""
        return CVE_FAIL_TPL.lstrip()

    @property
    def formatted_description(self) -> str:
        """Indented CVE description."""
        return '\n  '.join(textwrap.wrap(self.description))

    @property
    def id(self) -> str:
        """CVE id/code."""
        return self.cve_data["id"]

    @property
    def last_modified_date(self) -> date:
        """Date the CVE was last modified."""
        return self.parse_cve_date(self.cve_data['last_modified_date'])

    @property
    def published_date(self) -> date:
        """Date the CVE was published."""
        return self.parse_cve_date(self.cve_data['published_date'])

    @property
    def score(self) -> float:
        """CVE score."""
        return float(self.cve_data["score"])

    @property
    def severity(self) -> str:
        """CVE severity."""
        return self.cve_data["severity"]

    def format_failure(self, dep: "abstract.ADependency") -> str:
        """Format CVE failure for a given dependency."""
        return self.fail_template.render(
            cve=self,
            dependency=dep)

    def parse_cve_date(self, date_str: str) -> date:
        if not date_str.endswith('Z'):
            raise exceptions.CVEError(
                "CVE dates should be UTC and in isoformat")
        return date.fromisoformat(date_str.split('T')[0])



"""Abstract dependency checker."""

import abc
import argparse
import json
import math
import os
import pathlib
import re
from functools import cached_property
from typing import Optional, Tuple, Type

import aiohttp

import gidgethub

import abstracts

from aio.api import github as _github, nist
from aio.run import checker
from aio.core.tasks import ConcurrentError, inflate

from envoy.dependency.check import abstract, exceptions, typing


NO_GITHUB_TOKEN_ERROR_MSG = (
    "No Github access token supplied "
    "via environment variable `GITHUB_TOKEN` "
    "or argument `--github_token`")
NO_ISSUE_DEPENDENCIES = r"com_google_protobuf_protoc_[a-zA-Z0-9_]+$"


class ADependencyChecker(
        checker.Checker,
        metaclass=abstracts.Abstraction):
    """Dependency checker."""

    checks = (
        "cves",
        "release_dates",
        "release_issues",
        "releases")

    @property
    @abc.abstractmethod
    def access_token(self) -> Optional[str]:
        """Github access token."""
        if self.args.github_token:
            return pathlib.Path(self.args.github_token).read_text().strip()
        return os.getenv('GITHUB_TOKEN')

    @property
    def cve_config(self):
        return self.args.cve_config

    @cached_property
    def cves(self) -> "abstract.ADependencyCVEs":
        return self.cves_class(
            self.dependencies,
            config_path=self.cve_config,
            preloaded_cve_data=self.preloaded_cve_data,
            session=self.session,
            loop=self.loop,
            pool=self.pool)

    @property  # type:ignore
    @abstracts.interfacemethod
    def cves_class(self) -> Type["abstract.ADependencyCVEs"]:
        """CVEs class."""
        raise NotImplementedError

    @cached_property
    def dep_ids(self) -> Tuple[str, ...]:
        """Tuple of dependency ids."""
        return tuple(dep.id for dep in self.dependencies)

    @cached_property
    def dependencies(self) -> Tuple["abstract.ADependency", ...]:
        """Tuple of dependencies."""
        deps = []
        for k, v in self.dependency_metadata.items():
            deps.append(
                self.dependency_class(
                    k, v,
                    self.github,
                    pool=self.pool,
                    loop=self.loop))
        return tuple(sorted(deps))

    @property  # type:ignore
    @abstracts.interfacemethod
    def dependency_class(self) -> Type["abstract.ADependency"]:
        """Dependency class."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dependency_metadata(self) -> typing.DependenciesDict:
        """Dependency metadata (derived in Envoy's case from
        `repository_locations.bzl`)."""
        return json.loads(self.repository_locations_path.read_text())

    @cached_property
    def disabled_checks(self):
        disabled = {}
        if not self.access_token:
            disabled["release_dates"] = NO_GITHUB_TOKEN_ERROR_MSG
            disabled["release_issues"] = NO_GITHUB_TOKEN_ERROR_MSG
            disabled["releases"] = NO_GITHUB_TOKEN_ERROR_MSG
        return disabled

    @property
    def download_cves(self) -> str:
        return self.args.download_cves

    @cached_property
    def github(self) -> _github.IGithubAPI:
        """Github API."""
        return _github.GithubAPI(
            self.session, "",
            oauth_token=self.access_token)

    @cached_property
    def github_dependencies(self) -> Tuple["abstract.ADependency", ...]:
        """Tuple of dependencies."""
        deps = []
        for dep in self.dependencies:
            if not dep.github_url:
                urls = "\n".join(dep.urls)
                self.log.info(f"{dep.id} is not a GitHub repository\n{urls}")
                continue
            try:
                dep.github_version
            except exceptions.BadGithubURL as e:
                self.error("dependencies", [e.args[0]])
            else:
                deps.append(dep)
        return tuple(deps)

    @cached_property
    def issues(self) -> _github.IGithubIssuesTracker:
        """Dependency issues."""
        return self.issues_class(self.github)

    @property  # type:ignore
    @abstracts.interfacemethod
    def issues_class(self) -> Type[_github.IGithubIssuesTracker]:
        """Dependency issues class."""
        raise NotImplementedError

    @property
    def preloaded_cve_data(self) -> str:
        return self.args.cve_data

    @property
    def repository_locations_path(self) -> pathlib.Path:
        return pathlib.Path(self.args.repository_locations)

    @cached_property
    def session(self) -> aiohttp.ClientSession:
        """HTTP client session."""
        return aiohttp.ClientSession()

    @cached_property
    def sha_preload_limit(self) -> int:
        proc_count = os.cpu_count() or 1
        return math.ceil(proc_count * 1.5)

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument('--github_token')
        parser.add_argument('--repository_locations')
        parser.add_argument('--cve_config')
        parser.add_argument('--cve_data')
        parser.add_argument('--download_cves')

    async def check_cves(self) -> None:
        """Scan for CVEs in a parsed NIST CVE database."""
        for dep in self.dependencies:
            await self.dep_cve_check(dep)

    async def check_release_dates(self) -> None:
        """Check recorded dates match for dependencies."""
        for dep in self.github_dependencies:
            await self.dep_date_check(dep)

    async def check_release_issues(self) -> None:
        """Check dependency issues."""
        try:
            await self.release_issues_labels_check()
            for dep in self.github_dependencies:
                await self.dep_release_issue_check(dep)
            await self.release_issues_missing_dep_check()
            await self.release_issues_duplicate_check()
        except _github.exceptions.IssueCreateError as e:
            self.error(
                self.active_check,
                [f"Release issues check failed: {e}"])

    async def check_releases(self) -> None:
        """Check dependencies for new releases."""
        for dep in self.github_dependencies:
            await self.dep_release_check(dep)

    async def check_release_sha(self) -> None:
        """Check shas for new releases."""
        for dep in self.github_dependencies:
            await self.dep_release_sha_check(dep)

    async def dep_cve_check(
            self,
            dep: "abstract.ADependency") -> None:
        if not dep.cpe:
            self.log.info(f"No CPE listed for: {dep.id}")
            return
        warnings = []
        async for failing_cve in self.cves.dependency_check(dep):
            warnings.append(
                f'{failing_cve.format_failure(dep)}')
        if warnings:
            self.warn(
                self.active_check,
                warnings)
        else:
            self.succeed(
                self.active_check,
                [f"No CVE vulnerabilities found: {dep.id}"])

    async def dep_date_check(
            self,
            dep: "abstract.ADependency") -> None:
        """Check dates for dependency."""
        if not await dep.release.date:
            self.warn(
                self.active_check,
                [f"{dep.id} is a GitHub repository with no inferrable "
                 "release date"])
        elif await dep.release_date_mismatch:
            self.error(
                self.active_check,
                [f"Mismatch: {dep.id} "
                 f"{dep.release_date} != {await dep.release.date}"])
        else:
            self.succeed(
                self.active_check,
                [f"Match ({dep.release_date}): {dep.id}"])

    async def dep_release_issue_check(
            self,
            dep: "abstract.ADependency") -> None:
        """Check issues for dependency."""
        issue = (await self.issues["releases"].issues).get(dep.id)

        # TODO: move the exclusion logic to tracker
        if self._no_dep_issues.match(dep.id):
            if issue:
                # There is an open issue, but the dep shoudl be ignored.
                self.warn(
                    self.active_check,
                    [f"Incorrect issue: {dep.id} #{issue.number}"])
                if self.fix:
                    await self._dep_release_issue_close_stale(issue, dep)
            else:
                # No issue required
                self.log.info(
                    f"Ignored by depdendency issue tracker: {dep.id}")
            return

        if not (newer_release := await dep.newer_release):
            if issue:
                # There is an open issue, but the dep is already
                # up-to-date.
                self.warn(
                    self.active_check,
                    [f"Stale issue: {dep.id} #{issue.number}"])
                if self.fix:
                    await self._dep_release_issue_close_stale(issue, dep)
            else:
                # No issue required
                self.succeed(
                    self.active_check,
                    [f"No issue required: {dep.id}"])
            return
        if issue:
            if issue.version == (await dep.newer_release).version:
                # Required issue exists
                self.succeed(
                    self.active_check,
                    [f"Issue exists (#{issue.number}): {dep.id}"])
                return
            # Existing issue is showing incorrect version
            self.warn(
                self.active_check,
                [f"Out-of-date issue (#{issue.number}): {dep.id} "
                 f"({issue.version} -> {newer_release.version})"])
        else:
            # Issue is required to be added
            self.warn(
                self.active_check,
                [f"Missing issue: {dep.id} ({newer_release.version})"])
        if self.fix:
            await self._dep_release_issue_create(issue, dep)

    async def dep_release_check(
            self,
            dep: "abstract.ADependency") -> None:
        """Check releases for dependency."""
        if newer_release := await dep.newer_release:
            self.warn(
                self.active_check,
                [f"Newer release ({newer_release.tag_name}): {dep.id}\n"
                 f"{dep.release_date} "
                 f"{dep.github_version_name}\n"
                 f"{await newer_release.date} "
                 f"{newer_release.tag_name} "])
        elif await dep.has_recent_commits:
            self.warn(
                self.active_check,
                [f"Recent commits ({await dep.recent_commits}): {dep.id}\n"
                 f"There have been {await dep.recent_commits} commits since "
                 f"{dep.github_version_name} landed on "
                 f"{dep.release_date}"])
        else:
            self.succeed(
                self.active_check,
                [f"Up-to-date ({dep.github_version_name}): {dep.id}"])

    async def dep_release_sha_check(
            self,
            dep: "abstract.ADependency") -> None:
        """Check sha for dependency."""
        if not await dep.release.sha:
            self.error(
                self.active_check,
                [f"Unable to generate release SHA: {dep.id}"])
        elif await dep.release_sha_mismatch:
            self.error(
                self.active_check,
                [f"Mismatch: {dep.id} "
                 f"{dep.release_sha} != {await dep.release.sha}"])
        else:
            self.succeed(
                self.active_check,
                [f"Match ({dep.display_sha}): {dep.id}"])

    async def release_issues_duplicate_check(self) -> None:
        """Check for duplicate issues for dependencies."""
        duplicates = False
        async for issue in self.issues["releases"].duplicate_issues:
            duplicates = True
            self.warn(
                self.active_check,
                [f"Duplicate issue for dependency (#{issue.number}): "
                 f"{issue.key}"])
            if self.fix:
                await self._release_issue_close_duplicate(issue)
        if not duplicates:
            self.succeed(
                self.active_check,
                ["No duplicate issues found."])

    async def release_issues_labels_check(self) -> None:
        """Check expected labels are present."""
        missing = False
        for label in await self.issues["releases"].missing_labels:
            missing = True
            # TODO: make this a warning if `fix` and fix it
            self.error(
                self.active_check,
                [f"Missing label: {label}"])
        if not missing:
            self.succeed(
                self.active_check,
                [f"All ({len(self.issues['releases'].labels)}) "
                 "required labels are available."])

    async def release_issues_missing_dep_check(self) -> None:
        """Check for missing dependencies for issues."""
        closed = False
        issues = await self.issues["releases"].open_issues
        for issue in issues:
            if issue.key not in self.dep_ids:
                closed = True
                self.warn(
                    self.active_check,
                    [f"Missing dependency (#{issue.number}): {issue.key}"])
                if self.fix:
                    await self._release_issue_close_missing_dep(issue)
        if not closed:
            self.succeed(
                self.active_check,
                [f"All ({len(issues)}) issues have current dependencies."])

    async def on_checks_complete(self) -> int:
        await self.session.close()
        return await super().on_checks_complete()

    @checker.preload(
        when=["cves"],
        catches=[exceptions.CVECheckError, nist.exceptions.NISTError])
    async def preload_cves(self) -> None:
        async for download in self.cves.downloads:
            self.log.debug(f"Preloaded cve data: {download}")

    @checker.preload(
        when=["release_dates"],
        unless=["releases", "release_issues"],
        catches=[ConcurrentError, gidgethub.GitHubException])
    async def preload_release_dates(self) -> None:
        preloader = inflate(
            self.github_dependencies,
            lambda d: (
                d.release.date, ))
        async for dep in preloader:
            self.log.debug(f"Preloaded release date: {dep.id}")

    @checker.preload(
        when=["release_issues"],
        blocks=["release_dates"],
        catches=[gidgethub.GitHubException])
    async def preload_release_issues(self) -> None:
        await self.issues["releases"].missing_labels
        await self.issues["releases"].issues

    @checker.preload(
        when=["release_sha"],
        catches=[ConcurrentError, aiohttp.ClientError])
    async def preload_release_sha(self) -> None:
        preloader = inflate(
            self.github_dependencies,
            lambda d: (
                d.release.sha, ), limit=self.sha_preload_limit)
        async for dep in preloader:
            self.log.debug(f"Preloaded release sha: {dep.id}")

    @checker.preload(
        when=["releases", "release_issues"],
        blocks=["release_dates"],
        catches=[ConcurrentError, gidgethub.GitHubException])
    async def preload_releases(self) -> None:
        preloader = inflate(
            self.github_dependencies,
            lambda d: (
                d.newer_release,
                d.recent_commits))
        async for dep in preloader:
            self.log.debug(f"Preloaded release data: {dep.id}")

    async def run(self) -> Optional[int]:
        # TODO: figure out a better way for preloading data and saving it
        if self.download_cves:
            await self.cves.download_cves(self.download_cves)
            return 0
        return await super().run()

    @cached_property
    def _no_dep_issues(self):
        return re.compile(NO_ISSUE_DEPENDENCIES)

    async def _dep_release_issue_close_stale(
            self,
            issue: "abstract.AGithubDependencyReleaseIssue",
            dep: "abstract.ADependency") -> None:
        await issue.close()
        self.log.notice(
            f"Closed stale issue (#{issue.number}): {dep.id}\n"
            f"{issue.title}\n{issue.body}")

    async def _dep_release_issue_create(
            self,
            issue: "abstract.AGithubDependencyReleaseIssue",
            dep: "abstract.ADependency") -> None:
        if await self.issues["releases"].missing_labels:
            self.error(
                self.active_check,
                [f"Unable to create issue for {dep.id}: missing labels"])
            return
        new_issue = await self.issues["releases"].create(dep=dep)
        self.log.notice(
            f"Created issue (#{new_issue.number}): "
            f"{dep.id} {new_issue.version}\n"
            f"{new_issue.title}\n{new_issue.body}")
        if not issue:
            return
        await new_issue.close_old(issue, dep=dep)
        self.log.notice(
            f"Closed old issue (#{issue.number}): "
            f"{dep.id} {issue.version}\n"
            f"{issue.title}\n{issue.body}")

    async def _release_issue_close_duplicate(
            self,
            issue: "abstract.AGithubDependencyReleaseIssue") -> None:
        current_issue = (
            await self.issues["releases"].issues)[issue.key]
        await current_issue.close_duplicate(issue)
        self.log.notice(
            f"Closed duplicate issue (#{issue.number}): {issue.key}\n"
            f" {issue.title}\n"
            f"current issue #({current_issue.number}):\n"
            f" {current_issue.title}")

    async def _release_issue_close_missing_dep(
            self,
            issue: "abstract.AGithubDependencyReleaseIssue") -> None:
        """Close an issue that has no current dependency."""
        await issue.close()
        self.log.notice(
            f"Closed issue with no current dependency (#{issue.number})")




from functools import cached_property
from typing import Dict, List, Optional, Type

import abstracts

from aio.api import github, nist

from envoy.dependency import check


class DependencyCVE(check.ADependencyCVE):
    pass


class DependencyCVEs(check.ADependencyCVEs):

    @property
    def cpe_class(self) -> Type[nist.ACPE]:
        return nist.CPE

    @property
    def cve_class(self) -> Type[check.ADependencyCVE]:
        return DependencyCVE

    @cached_property
    def ignored_cves(self) -> List[str]:
        return super().ignored_cves

    @property
    def nist_downloader_class(self) -> Type[nist.NISTDownloader]:
        return nist.NISTDownloader


class Dependency(check.ADependency):

    @property
    def release_class(self) -> Type[check.ADependencyGithubRelease]:
        return DependencyGithubRelease


class DependencyGithubRelease(check.ADependencyGithubRelease):
    pass


class GithubDependencyReleaseIssue(check.AGithubDependencyReleaseIssue):
    pass


@abstracts.implementer(github.IGithubTrackedIssues)
class GithubDependencyReleaseIssues(check.AGithubDependencyReleaseIssues):

    @property
    def issue_class(self) -> Type[GithubDependencyReleaseIssue]:
        return GithubDependencyReleaseIssue


@abstracts.implementer(github.IGithubIssuesTracker)
class GithubDependencyIssuesTracker(github.AGithubIssuesTracker):

    @cached_property
    def tracked_issues(self) -> Dict:
        return dict(
            releases=GithubDependencyReleaseIssues(self.github))


class DependencyChecker(check.ADependencyChecker):

    @property
    def access_token(self) -> Optional[str]:
        return super().access_token

    @property
    def cves_class(self) -> Type[check.ADependencyCVEs]:
        return DependencyCVEs

    @property
    def dependency_class(self) -> Type[check.ADependency]:
        return Dependency

    @cached_property
    def dependency_metadata(self) -> check.typing.DependenciesDict:
        return super().dependency_metadata

    @property
    def issues_class(self) -> Type[github.IGithubIssuesTracker]:
        return GithubDependencyIssuesTracker




# Thrown on errors related to release date or version.
class ReleaseDateVersionError(Exception):
    pass


# Errors that happen during issue creation.
class DependencyUpdateError(Exception):
    pass


class BadGithubURL(Exception):
    pass


class NotGithubDependency(Exception):
    pass


class NoReleaseAssetError(Exception):
    pass


class CPEError(Exception):
    pass


class CVEError(Exception):
    pass


class CVECheckError(Exception):
    pass




from . import abstract, exceptions, typing
from .abstract import (
    ADependency,
    ADependencyChecker,
    ADependencyCVE,
    ADependencyCVEs,
    ADependencyGithubRelease,
    AGithubDependencyReleaseIssue,
    AGithubDependencyReleaseIssues)
from .checker import (
    Dependency,
    DependencyChecker,
    DependencyCVE,
    DependencyCVEs,
    DependencyGithubRelease,
    GithubDependencyIssuesTracker,
    GithubDependencyReleaseIssue,
    GithubDependencyReleaseIssues)
from .cmd import run, main
from . import checker


__all__ = (
    "abstract",
    "ADependency",
    "ADependencyChecker",
    "ADependencyCVE",
    "ADependencyCVEs",
    "ADependencyGithubRelease",
    "AGithubDependencyReleaseIssue",
    "AGithubDependencyReleaseIssues",
    "AGithubIssueManager",
    "checker",
    "Dependency",
    "DependencyChecker",
    "DependencyCVE",
    "DependencyCVEs",
    "DependencyGithubRelease",
    "GithubDependencyIssuesTracker",
    "GithubDependencyReleaseIssue",
    "GithubDependencyReleaseIssues",
    "exceptions",
    "main",
    "run",
    "typing")





from typing import Dict, List, Optional, TypedDict

from aio.api import nist


# Scanner configuration

class CVEConfigDict(TypedDict, total=False):
    ignored_cves: List[str]
    nist_url: str
    start_year: int


class DependencyCVEItemDict(nist.typing.CVEItemDict, total=False):
    score: str
    description: str
    last_modified_date: str
    published_date: str
    cpes: nist.typing.CPEsTuple
    severity: str


# Package defined types

class BaseDependencyMetadataDict(TypedDict):
    release_date: str
    version: str


class DependencyMetadataDict(BaseDependencyMetadataDict, total=False):
    cpe: Optional[str]
    urls: List[str]
    sha256: str


DependenciesDict = Dict[str, DependencyMetadataDict]



"""Abstract dependency."""

import asyncio
import logging
from concurrent import futures
from functools import cached_property
from typing import List, Optional, Set, Type

import gidgethub

from packaging import version

import abstracts

from aio.api import github
from aio.core import event
from aio.core.functional import async_property
from aio.core.tasks import ConcurrentError

from envoy.dependency.check import abstract, exceptions, typing

logger = logging.getLogger(__name__)


@abstracts.implementer(event.IReactive)
class ADependency(event.AReactive, metaclass=abstracts.Abstraction):
    """Github dependency."""

    def __init__(
            self,
            id: str,
            metadata: "typing.DependencyMetadataDict",
            github: github.IGithubAPI,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        self.id = id
        self.metadata = metadata
        self.github = github
        self._loop = loop
        self._pool = pool

    def __gt__(self, other: "ADependency") -> bool:
        return self.id > other.id

    def __lt__(self, other: "ADependency") -> bool:
        return self.id < other.id

    def __str__(self):
        return f"{self.id}@{self.display_version}"

    @async_property(cache=True)
    async def commits_since_current(self) -> int:
        """Commits since current commit/tag."""
        count = await self.repo.commits(
            since=await self.release.timestamp_commit).total_count
        return count and count - 1 or count

    @cached_property
    def cpe(self) -> Optional[str]:
        """Configured CPE for this dependency."""
        return (
            str(self.metadata["cpe"])
            if self.metadata.get("cpe", "N/A") != "N/A"
            else None)

    @cached_property
    def display_sha(self) -> str:
        """Truncated release sha of this dependency, may/not be the same as the
        `display_version`."""
        return self.release_sha[:10]

    @cached_property
    def display_version(self) -> str:
        """Version string of this dependency, truncated if necessary."""
        return (
            self.version[:10]
            if not self.release.tagged
            else self.version)

    @property
    def github_filetypes(self) -> Set[str]:
        return {".tar.gz", ".zip"}

    @cached_property
    def github_url(self) -> str:
        """Github URL."""
        for url in self.urls:
            if url.startswith('https://github.com/'):
                return url
        return ""

    @cached_property
    def github_version(self) -> str:
        """Github version, as parsed from the URL."""
        if self.url_components[5] != 'archive':
            # Release tag is a path component.
            if self.url_components[5] != 'releases':
                raise exceptions.BadGithubURL(
                    "Unable to parse github URL components: "
                    f"{self.url_components[3:]}")
            return self.url_components[7]
        # Only support .tar.gz, .zip today. Figure out the release tag
        # from this filename.
        for filetype in self.github_filetypes:
            if self.url_components[-1].endswith(filetype):
                return self.url_components[-1][:-len(filetype)]
        raise exceptions.BadGithubURL(
            "Unrecognized Github release asset: "
            f"{self.url_components[3:]}")

    @property
    def github_version_name(self) -> str:
        """Github version, truncated to 7 char if its sha_hash."""
        return (
            self.github_version[0:7]
            if not self.release.tagged
            else self.github_version)

    @async_property
    async def has_recent_commits(self) -> bool:
        """Flag indicating whether there are more recent commits than the
        current pinned commit."""
        try:
            return await self.recent_commits > 1
        except (ConcurrentError, gidgethub.GitHubException) as e:
            logger.debug(
                f"Fetching recent commits failed ({self}): {type(e)} {e}")
            raise e

    @async_property(cache=True)
    async def newer_release(
            self) -> Optional["abstract.ADependencyGithubRelease"]:
        """Release with highest semantic version if newer than the current
        release, or where pin is to tag or commit."""
        # TODO: consider adding `newer_tags` for deps that only create
        #   tags and not releases (eg tclap)
        try:
            newer_release = await self.repo.highest_release(
                since=await self.release.timestamp)
        except (ConcurrentError, gidgethub.GitHubException) as e:
            logger.debug(
                f"Fetching newer release failed ({self}): {type(e)} {e}")
            raise e
        return (
            self.release_class(
                self.repo,
                newer_release.tag_version,
                release=newer_release)  # type:ignore
            if (newer_release
                and (version.parse(newer_release.tag_version)
                     != self.release.version))
            else None)

    @property
    def organization(self) -> str:
        """Github organization name."""
        return self.url_components[3]

    @property
    def project(self) -> str:
        """Github project name."""
        return self.url_components[4]

    @async_property(cache=True)
    async def recent_commits(self) -> int:
        """Count of commits since current pinned commit."""
        return (
            await self.commits_since_current
            if not self.release.tagged
            else 0)

    @cached_property
    def release(self) -> "abstract.ADependencyGithubRelease":
        """Github release."""
        return self.release_class(
            self.repo,
            self.github_version,
            asset_url=self.github_url,
            loop=self.loop,
            pool=self.pool)

    @property  # type:ignore
    @abstracts.interfacemethod
    def release_class(self) -> Type["abstract.ADependencyGithubRelease"]:
        """Github release class."""
        raise NotImplementedError

    @property
    def release_date(self) -> str:
        """Release (or published) date of this dependency."""
        return self.metadata["release_date"]

    @async_property
    async def release_date_mismatch(self) -> bool:
        """Flag indicating the metadata date doesnt match the Github date."""
        return (
            self.release_date
            != await self.release.date)

    @property
    def release_sha(self) -> str:
        """Release (or published) sha of this dependency."""
        return self.metadata["sha256"]

    @async_property
    async def release_sha_mismatch(self) -> bool:
        """Flag indicating the metadata sha doesnt match the Github sha."""
        return (
            self.release_sha
            != await self.release.sha)

    @cached_property
    def release_version(self) -> Optional[version.Version]:
        """Semantic version for the release of this dependency if available."""
        try:
            return version.Version(self.version)
        except version.InvalidVersion:
            return None

    @cached_property
    def repo(self) -> github.IGithubRepo:
        """Github repo for this dependency."""
        return self.github[
            f"{self.organization}/{self.project}"]  # type: ignore

    @cached_property
    def url_components(self) -> List[str]:
        """Github URL components."""
        if not self.github_url:
            urls = "\n".join(self.urls)
            raise exceptions.NotGithubDependency(
                f'{self.id} is not a GitHub repository\n{urls}')
        # TODO: add/use a proper `GithubURLParser`
        return self.github_url.split('/')

    @property
    def urls(self) -> List[str]:
        """Urls of this dependency."""
        return self.metadata["urls"]

    @property
    def version(self) -> str:
        """Version string of this dependency."""
        return self.metadata["version"]
